use crate::irust::{IRust, Result};
use crossterm::style::Color;
use irust_repl::{CompileMode, DEFAULT_EVALUATOR, Edition, Executor, MainResult, ToolChain};
use serde::{Deserialize, Serialize};
use std::io::{Read, Write};

#[derive(Deserialize, Serialize, Clone, Debug)]
#[serde(default)]
pub struct Options {
    add_irust_cmd_to_history: bool,
    add_shell_cmd_to_history: bool,
    pub ok_color: Color,
    pub eval_color: Color,
    pub irust_color: Color,
    pub irust_warn_color: Color,
    pub out_color: Color,
    pub shell_color: Color,
    pub err_color: Color,
    pub input_color: Color,
    pub insert_color: Color,
    pub welcome_msg: String,
    pub welcome_color: Color,
    pub ra_inline_suggestion_color: Color,
    pub ra_suggestions_table_color: Color,
    pub ra_selected_suggestion_color: Color,
    pub ra_max_suggestions: usize,
    pub first_irust_run: bool,
    pub enable_rust_analyzer: bool,
    pub toolchain: ToolChain,
    pub check_statements: bool,
    pub auto_insert_semicolon: bool,
    pub replace_marker: String,
    pub replace_output_with_marker: bool,
    pub input_prompt: String,
    pub output_prompt: String,
    pub activate_scripting: bool,
    pub executor: Executor,
    pub evaluator: Vec<String>,
    pub compile_time: bool,
    pub main_result: MainResult,
    pub show_warnings: bool,
    pub edition: Edition,
    pub debugger: Debugger,
    pub shell_interpolate: bool,
    pub local_server: bool,
    pub local_server_adress: std::net::SocketAddrV4,
    pub theme: String,
    pub compile_mode: CompileMode,
    pub new_lines_after_output: usize,

    #[serde(skip)]
    config_load_time: Option<std::time::SystemTime>,
}

impl Default for Options {
    fn default() -> Self {
        Self {
            // [Histroy]
            add_irust_cmd_to_history: true,
            add_shell_cmd_to_history: false,

            // [Colors]
            ok_color: Color::Blue,
            eval_color: Color::White,
            irust_color: Color::DarkBlue,
            irust_warn_color: Color::Cyan,
            out_color: Color::Red,
            shell_color: Color::DarkYellow,
            err_color: Color::DarkRed,
            input_color: Color::Yellow,
            insert_color: Color::White,

            // [Welcome]
            welcome_msg: String::new(),
            welcome_color: Color::DarkBlue,

            // [Rust Analyzer]
            enable_rust_analyzer: false,
            ra_inline_suggestion_color: Color::Cyan,
            ra_suggestions_table_color: Color::Green,
            ra_selected_suggestion_color: Color::DarkRed,
            ra_max_suggestions: 5,

            //other
            first_irust_run: true,
            toolchain: ToolChain::Default,
            check_statements: true,
            auto_insert_semicolon: true,

            // replace output
            replace_marker: "$out".into(),
            replace_output_with_marker: false,

            input_prompt: "In: ".to_string(),
            output_prompt: "Out: ".to_string(),
            activate_scripting: false,
            executor: Executor::Sync,
            evaluator: DEFAULT_EVALUATOR
                .iter()
                .map(|part| part.to_string())
                .collect(),
            compile_time: false,
            main_result: MainResult::Unit,
            show_warnings: false,
            edition: Edition::E2021,
            debugger: Debugger::LLDB,
            shell_interpolate: true,
            local_server: false,
            local_server_adress: "127.0.0.1:9000".parse().expect("correct"),
            theme: "default".into(),
            compile_mode: CompileMode::Debug,
            new_lines_after_output: 1,
            config_load_time: None,
        }
    }
}

impl Options {
    pub fn save(&mut self) -> Result<()> {
        if let Some(path) = Self::config_path() {
            // Check if the file has been modified since we loaded it
            if let Ok(metadata) = std::fs::metadata(&path) {
                if let Ok(modified_time) = metadata.modified() {
                    // If the config file has been modified since startup, don't overwrite it
                    if let Some(load_time) = self.config_load_time {
                        if modified_time > load_time {
                            // File was modified after we loaded it, don't overwrite
                            return Ok(());
                        }
                    }
                }
            }
            Self::write_config_file(path, self)?;
        }
        Ok(())
    }

    pub fn new() -> Result<Self> {
        if let Some(config_path) = Options::config_path() {
            let mut config_file = std::fs::File::open(&config_path)?;
            let mut config_data = String::new();
            config_file.read_to_string(&mut config_data)?;

            // Get the file's last modified time
            let modified_time = std::fs::metadata(&config_path)
                .ok()
                .and_then(|m| m.modified().ok());

            // Parse the config and store the load time
            let mut options: Options = toml::from_str(&config_data)?;
            options.config_load_time = modified_time;

            Ok(options)
        } else {
            Ok(Options::default())
        }
    }

    pub fn reset(&mut self) {
        *self = Self::default();
    }

    pub fn reset_evaluator(&mut self) {
        self.evaluator = DEFAULT_EVALUATOR
            .iter()
            .map(|part| part.to_string())
            .collect();
    }

    pub fn config_path() -> Option<std::path::PathBuf> {
        let config_dir = match crate::utils::irust_dirs::config_dir() {
            Some(dir) => dir.join("irust"),
            None => return None,
        };

        // Ignore directory exists error
        let _ = std::fs::create_dir_all(&config_dir);
        let config_path = config_dir.join("config.toml");

        Some(config_path)
    }

    fn write_config_file(config_path: std::path::PathBuf, options: &Options) -> Result<()> {
        let config = toml::to_string(options)?;

        let mut config_file = std::fs::File::create(config_path)?;

        write!(config_file, "{config}")?;
        Ok(())
    }
}

impl IRust {
    pub fn should_push_to_history(&self, buffer: &str) -> bool {
        let buffer: Vec<char> = buffer.chars().collect();

        if buffer.is_empty() {
            return false;
        }
        if buffer.len() == 1 {
            return buffer[0] != ':';
        }

        let irust_cmd = buffer[0] == ':' && buffer[1] != ':';
        let shell_cmd = buffer[0] == ':' && buffer[1] == ':';

        (irust_cmd && self.options.add_irust_cmd_to_history)
            || (shell_cmd && self.options.add_shell_cmd_to_history)
            || (!irust_cmd && !shell_cmd)
    }
    pub fn dont_save_options(&mut self) {
        self.engine.dont_save_options = true;
    }
}

#[allow(clippy::upper_case_acronyms)]
#[derive(Deserialize, Serialize, Clone, Debug)]
pub enum Debugger {
    LLDB,
    GDB,
}
